#include "mainNode/MainNode.h" 

#include "exceptions/Exceptions.h"

#include <pluginlib/class_list_macros.h>

#include <iostream>

#include <fstream>
#include <sstream>
#include <vector>
#include <iterator>

PLUGINLIB_EXPORT_CLASS(car::MainNode, nodelet::Nodelet);

namespace car
{

// TODO: ConfigReader!

platoonProtocol::VehicleId MainNode::readVehicleId()
{
    // open config file
    std::string userHome = getenv("HOME");
    std::ifstream configFile;
    std::string filename = userHome + "/CarConfig/platoon.config";
    configFile.open(filename, std::ifstream::in);
    if (!configFile.is_open())
        throw FileNotFound{filename};

    // desired parameters
    platoonProtocol::VehicleId vehicleId;
    bool foundVehicleId = false;

    std::string contentLine;
    while (!configFile.eof())
    {
        std::getline(configFile, contentLine);

        // split this line
        std::istringstream iss(contentLine);
        std::vector<std::string> words{std::istream_iterator<std::string>(iss), {} };

        if (words.size() > 1)
        {
            if (words.at(0) == "VehicleId:")
            {
                vehicleId = std::stoi(words.at(1));
                foundVehicleId = true;
            }
        }
    }

    if(!foundVehicleId)
        throw MissingParameters();

    configFile.close();

    messageOStream.write("vehicleId", vehicleId);
    return vehicleId;
}

platoonProtocol::Vehicle::VehicleEndpoints MainNode::readVehicleEndpoints()
{
    // open config file
    std::string userHome = getenv("HOME");
    std::ifstream configFile;
    std::string filename = userHome + "/CarConfig/platoon.config";
    configFile.open(filename, std::ifstream::in);
    if (!configFile.is_open())
        throw FileNotFound{filename};

    // desired parameters
    platoonProtocol::Vehicle::VehicleEndpoints endpoints;

    std::string contentLine;
    while (!configFile.eof())
    {
        std::getline(configFile, contentLine);

        // split this line
        std::istringstream iss(contentLine);
        std::vector<std::string> words{std::istream_iterator<std::string>(iss), {} };
        if (words.size() > 1)
        {
            if (words.at(0) == "VehicleEndpoint:")
            {
                auto vehicleId = static_cast<platoonProtocol::VehicleId>(std::stoi(words.at(1)));
                auto host = words.at(2);
                endpoints.emplace_back(platoonProtocol::VehicleEndpoint{vehicleId, host});
            }
        }
    }

    configFile.close();

    for (const auto & endpoint : endpoints)
    {
        messageOStream << MessageOStream::key << "VehicleEndpoint"
                       << MessageOStream::value << endpoint.getVehicleId() << " " << endpoint.getHost()
                       << MessageOStream::flush;
    }

    return endpoints;
}

MainNode::MainNode()
    : nh()
      , messageOStream(nh, "MainNode")
      , c2cNet()
      , c2c(platoonProtocol::Vehicle::Role::FOLLOWER, c2cNet, readVehicleId())
      , pcNet()
      , pc(pc2car::CommandReceiver::create(pcNet))
      , egoMotion(nh)
      , platoonController(nh, c2c, pc, egoMotion)
      , platoonControllerThread([this] { platoonController.run(); })
      , cruiseController(nh, c2c, egoMotion, platoonController)
      , cruiseControllerThread([this] { cruiseController.run(); })
{
    egoMotion.setCruiseControllerNotify(cruiseControllerNotify);
    egoMotion.setPlatoonControllerNotify(platoonControllerNotify);
    platoonController.setCruiseControllerNotify(cruiseControllerNotify);
    pc->receiveCommands([this] (auto commandCode) {platoonControllerThread.notify();} );
    c2c.setCallback( [this] {platoonControllerThread.notify();} );
    c2c.addVehicleEndpoints(readVehicleEndpoints());
}

void MainNode::onInit()
{
	messageOStream.write("onInit", "START");
    messageOStream.write("onInit", "END");
}

}
